#!/usr/bin/python3

import argparse
import requests
import random
import urllib
import json
import hashlib
import re
import time


DEFAULT_APTLY_URI = 'http://your-aptly-server:8080'


def log_msg(msg):
    print(f"{time.time()}: {msg}")


def md5(filename):
    hash_md5 = hashlib.md5()
    with open(filename, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()


def get_described_files(filename, check_md5):
    source_files = []

    with open(filename, 'r') as f:
        content = f.read()

    # Find the 'Files:' section in the dsc file and all lines starting with a space that follow it
    FILES_KEY = 'Files'
    match = re.search(r"Files:\s?\n(?P<{files}>(?:^ .*\n)+)".format(files=FILES_KEY), content, flags=re.MULTILINE)
    if not match:
        raise ValueError("Could not parse '{}' file".format(file))
    files = match.group(FILES_KEY)

    MD5SUM_KEY = 'Md5sum'
    SIZE_KEY = 'Size'
    FILENAME_KEY = 'Filename'
    SECTION_KEY = 'Section'
    PRIORITY_KEY = 'Priority'

    for file_match in re.finditer(
            r' (?P<{md5sum}>[0-9a-fA-F]+) (?P<{size}>[0-9]+) ?(?P<{section}>[0-9a-zA-Z-/]*) ?(?P<{priority}>[0-9a-zA-Z]*) (?P<{filename}>[^\s]+)$'.format(
                    md5sum=MD5SUM_KEY, size=SIZE_KEY, filename=FILENAME_KEY, section=SECTION_KEY,
                    priority=PRIORITY_KEY), files, flags=re.MULTILINE):
        filename = file_match.group(FILENAME_KEY)
        md5sum = file_match.group(MD5SUM_KEY)
        if check_md5 and md5(filename) != md5sum:
            raise ValueError("Md5sum of {} does not match its expected checksum".format(filename))

        source_files.append(filename)
    return source_files


def get_source_files(dsc_file, check_md5=True):
    return get_described_files(dsc_file, check_md5)


def get_binary_files(changes_file, check_md5=True):
    return get_described_files(changes_file, check_md5)


def check_error(reply, action):
    if 200 <= reply.status_code < 300:
        print(f'{action} succeeded')
    else:
        raise RuntimeError('{} failed:\n error code: {}\n error message: {}'
                           .format(action, reply.status_code, reply.text))


def sanitized_path(path):
    """ 
        Path arguments in aptly API calls, need to have all / characters replaced by underscores, 
        and all real underscores, by double underscores.
    """
    return path.replace('_', '__').replace('/', '_')


def random_name(prefix):
    r = random.SystemRandom()
    result = prefix
    for i in range(0, 8):
        result += r.choice('abcdefghijklmnopqrstuvwxyz')
    return result


def upload_packages(repo_name, distribution, path, aptly_uri, packages):
    tmp_dir_name = random_name('upload_')
    upload_uri = '{}/api/files/{}'.format(aptly_uri, tmp_dir_name)
    try:
        for pkg in packages:
            log_msg(f"uploading {pkg}...")
            with open(pkg, 'rb') as pkg_file:
                r = requests.post(upload_uri, files={
                    'file': pkg_file
                })
                check_error(r, 'Uploading file {}'.format(pkg))

        log_msg(f"adding the files to repo '{repo_name}'...")
        r = requests.post('{}/api/repos/{}/file/{}'.format(aptly_uri, repo_name, tmp_dir_name), json={
            'noRemove': 0,
            'forceReplace': 1
        })
        check_error(r, f'Adding files to repository {repo_name}')

        log_msg("publishing")
        r = requests.put('{}/api/publish/{}/{}'.format(aptly_uri, sanitized_path(path), distribution), json={
            'ForceOverwrite': True
        })
        check_error(r, 'Republishing repository {}'.format(repo_name))
    finally:
        # clean up the upload directory and if an exception occurred, re-raise it
        requests.delete(upload_uri)


def get_source_package_name(dsc_file):
    return dsc_file.split('_')[0]


def main():
    parser = argparse.ArgumentParser(description='Publish built packages. Working directory must be where the files are.')
    parser.add_argument('repo', metavar='REPO', type=str, help='Repository to publish to.')
    parser.add_argument('distribution', metavar='DISTRIBUTION', type=str, help='Apt distribution to publish to')
    parser.add_argument('dsc_file', metavar='DSC_FILE', type=str, help='The dsc file')
    parser.add_argument('changes', metavar='CHANGES', type=str,
                        help='The changes file containing the information on the packages to publish', nargs='+')
    parser.add_argument('-a', type=str,
                        help='URL of the aptly api server (default: {})'.format(DEFAULT_APTLY_URI),
                        dest='aptly_uri', default=DEFAULT_APTLY_URI)
    parser.add_argument('-p', type=str,
                        help='Path of the repo to publish (default: same as repo name with any "_" replaced by "__" and "/" replaced by "_")'.format(DEFAULT_APTLY_URI),
                        dest='repo_path')
    parser.add_argument('-s', '--source-package', default=None,
                        help='The source package name of the package. By default it is guessed from the dsc file name')

    args = parser.parse_args()

    packages = [args.dsc_file]
    packages.extend(get_source_files(args.dsc_file))
    for changes_file in args.changes:
        packages.append(changes_file)
        packages.extend(get_binary_files(changes_file))

    repo_path = args.repo_path if args.repo_path else args.repo
    print(f"Uploading the following files to repo '{args.repo}' at path '{repo_path}':")
    for package in packages:
        print(" {}".format(package))

    upload_packages(args.repo, args.distribution, args.repo, args.aptly_uri, packages)


main()

